// ---------------------------------------------------------------------------
//
//  This file is part of PermLib.
//
// Copyright (c) 2009-2011 Thomas Rehn <thomas@carmen76.de>
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. The name of the author may not be used to endorse or promote products
//    derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
// OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
// NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// ---------------------------------------------------------------------------


#ifndef TEST_COMMON_H_
#define TEST_COMMON_H_

#include <list>
#include <boost/shared_ptr.hpp>

#include <permlib/common.h>
#include <permlib/bsgs.h>
#include <permlib/construct/schreier_sims_construction.h>
#include <permlib/construct/random_schreier_sims_construction.h>
#include <permlib/construct/cyclic_group_construction.h>
#include <permlib/generator/product_replacement_generator.h>

#include <iostream>
#include <fstream>

namespace permlib { namespace test {

struct GroupInformation {
    std::string filename;
    unsigned int n;
    unsigned int order;
    
    GroupInformation(const std::string &f, unsigned int _n, unsigned int _order) : filename(f), n(_n), order(_order) {}
};

const GroupInformation info_trivial("../data/G_trivial42", 42, 1);
const GroupInformation info_test33("../data/G_test33", 33, 163680);
const GroupInformation info_psu4_3("../data/G_psu4_3", 820, 3265920);
const GroupInformation info_1997("../data/G_test1997", 13, 13*12*9*4);
const GroupInformation info_e6("../data/G_e6", 36, 51840);
const GroupInformation info_e7("../data/G_e7", 63, 1451520);
const GroupInformation info_e8("../data/G_e8", 120, 348364800);
const GroupInformation info_cyclic500("../data/G_cyclic500", 500, 500);
const GroupInformation info_cyclic10("../data/G_cyclic10", 10, 10);
const GroupInformation info_cyclic37("../data/G_cyclic37", 37, 37);
const GroupInformation info_cyclic37_2("../data/G_cyclic37-2", 74, 37);
const GroupInformation info_myciel3("../data/G_myciel3", 33, 60);
const GroupInformation info_cov1075("../data/G_cov1075", 120, 3628800L);
// wreath product of S_3 and S_5
const GroupInformation info_S3wrS5("../data/G_S3wrS5", 15, 933120L);
// wreath product of S_5 and S_3
const GroupInformation info_S5wrS3("../data/G_S5wrS3", 15, 10368000L);
const GroupInformation info_S6("../data/G_S6", 6, 720L);
const GroupInformation info_S6_3("../data/G_S6-3", 18, 720L);
// Klein four-group ~ C_2 x C_2
const GroupInformation info_Klein4("../data/G_klein4", 4, 4L);

template<class PERM>
unsigned int readGroup(const std::string &filename, std::list<typename PERM::ptr> &groupGenerators) {
    std::ifstream file;
    file.open(filename.c_str());
	if (!file.is_open()) {
		std::cerr << "opening " << filename << " failed" << std::endl;
        BOOST_REQUIRE(file.is_open());
		return 0;
	}
    
    unsigned int n;
    file >> n;

    std::string line;
    while (!file.eof()) {
        std::getline(file, line);
        if (line.length() < 1)
            continue;
		boost::shared_ptr<PERM> gen(new PERM(n, line));
        groupGenerators.push_back(gen);
    }
    file.close();
    
    return n;
}

using namespace permlib;

template<class PERM,class TRANS>
struct Constructor {
	typedef PERM PERMtype;
	typedef TRANS TRANStype;
};

template<class PERM,class TRANS>
struct ConstructDeterministic1 : public Constructor<PERM,TRANS> {
	BSGS<PERM, TRANS> operator()(const GroupInformation& info) {
		std::list<typename PERM::ptr> groupGenerators;
		const unsigned int n = readGroup<PERM>(info.filename, groupGenerators);

		SchreierSimsConstruction<PERM, TRANS> ssc(n);
		BSGS<PERM, TRANS> bsgs = ssc.construct(groupGenerators.begin(), groupGenerators.end());
		
		return bsgs;
	}
};

template<class PERM,class TRANS>
struct ConstructDeterministic2 : public Constructor<PERM,TRANS> {
	BSGS<PERM, TRANS> operator()(const GroupInformation& info) {
		std::list<typename PERM::ptr> groupGenerators;
		const unsigned int n = readGroup<PERM>(info.filename, groupGenerators);

		SchreierSimsConstruction<PERM, TRANS> ssc(n);
		BSGS<PERM, TRANS> bsgs(n);
		bsgs = ssc.construct(groupGenerators.begin(), groupGenerators.end());
		
		return bsgs;
	}
};

template<class PERM,class TRANS>
struct ConstructRandomized : public Constructor<PERM,TRANS> {
	BSGS<PERM, TRANS> operator()(const GroupInformation& info) {
		std::list<typename PERM::ptr> groupGenerators;
		const unsigned int n = readGroup<PERM>(info.filename, groupGenerators);

		ProductReplacementGenerator<PERM> rng(n, groupGenerators.begin(), groupGenerators.end());
		RandomSchreierSimsConstruction<PERM, TRANS> ssc(n, &rng, info.order);
		bool result = false;
		BSGS<PERM, TRANS> bsgs = ssc.construct(groupGenerators.begin(), groupGenerators.end(), result);
		BOOST_REQUIRE(result);
		
		return bsgs;
	}
};

template<class PERM,class TRANS>
BSGS<PERM, TRANS> construct(const GroupInformation& info) {
	return ConstructDeterministic1<PERM,TRANS>()(info);
}

template<class PERM, class TRANS>
BSGS<PERM, TRANS> constructCyclicGroup(unsigned int n) {
	CyclicGroupConstruction<TRANS> cycConstruct(n);
	return cycConstruct.construct();
}

template<class PERM>
void checkImage(const unsigned long* Delta, unsigned long DeltaSize, const unsigned long* Phi, const boost::shared_ptr<PERM> &repr) {
	std::list<unsigned long> imgList(Phi, Phi+DeltaSize);
	for (unsigned int j = 0; j < DeltaSize; ++j) {
		std::list<unsigned long>::iterator it = std::find(imgList.begin(), imgList.end(), *repr / Delta[j]);
		BOOST_REQUIRE( it != imgList.end() );
		imgList.erase(it);
	}
	BOOST_CHECK( imgList.size() == 0 );
}

} } // end NS

#endif // - TEST_COMMON_H_
